from django.core.files.storage import get_storage_class


class SvgFile:
    """
    Object to mimic Sorl object ``thumbnail.images.ImageFile`` and implement some of
    its attributes and methods.

    All other method about image size are not supported.

    Attributes:
        name (string): File name, aka its relative path from its storage.
        url (string): Full media file url from its storage.

    Arguments:
        fileobject (django.core.files.File): A valid Django file object.

    Keyword Arguments:
        storage (django.core.files.storage.FileSystemStorage): Storage related
            to the file. If not given, the default Django storage backend is
            used.
    """
    def __init__(self, fileobject, storage=None):
        self.name = fileobject.name
        self.storage = storage or get_storage_class()()

    def __str__(self):
        return self.name

    def exists(self):
        return self.storage.exists(self.name)

    @property
    def url(self):
        return self.storage.url(self.name)



"""
Default application settings
----------------------------

These are the default settings you can override in your own project settings
right after the line which load the default app settings.

"""
SMARTIMAGE_FILESIZE_LIMIT = 5242880
"""
Maximum file size allowed for upload.

This is a limit at Django level so files are still stored until post validation.

You should mind to configure a limit on your webserver to avoid basic attacks
with very big files.

Value could be something like:

* ``5242880`` for ~5MiO;
* ``10485760`` for ~10MiO;
* ``42991616`` for ~50MiO.

See:

https://stackoverflow.com/questions/2472422/django-file-upload-size-limit
"""

SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS = [
    "jpg",
    "jpeg",
    "svg",
    "gif",
    "png",
]
"""
Allowed image file extensions for image files.

You should accord this list with setting ``SMART_FORMAT_AVAILABLE_FORMATS``.

This is used by ``SmartMediaField`` which includes ``SmartMediaFileExtensionValidator``
that use this setting.

However, you can also use it yourself to your own fields with Django validator
``FileExtensionValidator``: ::

    class MyField(FileField):
        ...
        default_validators = [
            FileExtensionValidator(
                allowed_extensions=settings.SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS
            ),
        ]

However this won't work with ImageField since allowed extensions contains SVG that is
not a supported image by PIL.

"""

SMART_FORMAT_AVAILABLE_FORMATS = [
    ("jpg", "JPEG"),
    ("jpeg", "JPEG"),
    ("png", "PNG"),
    ("gif", "GIF"),
    ("svg", "SVG"),
]
"""
Available formats for template tag ``media_thumb``.

This is a list of tuples such as ``(extension, name)``, where ``extension`` is
a lowercase file extension (without leading dot) and ``name`` is the format name as
expected from Sorl (excepted for SVG which does not goes through Sorl).

You should accord this list with setting ``SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS``. All
file fields that are used with template tag ``media_thumb`` must have a format
registered in this setting.
"""



from django.conf import settings
from django.core.validators import FileExtensionValidator
from django.forms import ValidationError
from django.template.defaultfilters import filesizeformat
from django.utils.translation import gettext_lazy as _


def validate_file_size(data):
    """
    Validate file size does not exceed limit from
    ``settings.SMARTIMAGE_FILESIZE_LIMIT``.

    Raises:
        ValidationError: If file size is over limit.

    Arguments:
        data (file object):

    Returns:
        boolean: Always True, obvisously excepted if an exception is raised
        when file is over limit.
    """
    msg = _('Please keep file size under {}. Your one was {}')

    if data.size > settings.SMARTIMAGE_FILESIZE_LIMIT:
        raise ValidationError(msg.format(
            filesizeformat(settings.SMARTIMAGE_FILESIZE_LIMIT),
            filesizeformat(data.size)
        ))

    return True


class SmartMediaFileExtensionValidator(FileExtensionValidator):
    """
    Customized ``FileExtensionValidator`` to directly get the allowed extension from
    setting ``SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS``.

    Since allowed extensions are embedded in code from settings, Django won't detect
    any change model change when changing the setting and so won't create a migration.

    Argument ``allowed_extensions`` is still accepted but is ignored.
    """
    def __init__(self, *args, **kwargs):
        kwargs["allowed_extensions"] = settings.SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS

        super().__init__(*args, **kwargs)



from pathlib import Path

from django.contrib.sites.models import Site
from django.template.response import TemplateResponse
from django.test.html import parse_html
from django.urls import reverse

from pyquery import PyQuery as pq

from .factories import create_image_file


# A dummy password that should pass form validation
VALID_PASSWORD_SAMPLE = "Azerty12345678"

# This is the common dummy URL which Django REST Framework will use when it does not
# have any request when it resolve URL to absolute (like from Hyperlinked classes)
DRF_DUMMY_HOST_URL = "http://testserver"


# A dummy blank GIF file in byte value to simulate an uploaded file like with
# 'django.core.files.uploadedfile.SimpleUploadedFile'
DUMMY_GIF_BYTES = (
    b'\x47\x49\x46\x38\x39\x61\x01\x00\x01\x00\x00\x00\x00\x21\xf9\x04'
    b'\x01\x0a\x00\x01\x00\x2c\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02'
    b'\x02\x4c\x01\x00\x3b'
)


def get_website_url(site_settings=None):
    """
    A shortand to retrieve the full website URL according to Site ID and HTTP
    protocole settings.

    Keyword Arguments:
        site_settings (django.conf.settings): Settings object, if not given the method
            will be unable to determine if HTTPS is enabled or not, so it will always
            return a HTTP URL.

    Returns:
        string: Full website URL.
    """
    domain = Site.objects.get_current().domain

    protocol = "http://"
    if site_settings and site_settings.HTTPS_ENABLED:
        protocol = "https://"

    return "{}{}".format(protocol, domain)


def get_relative_path(site_url, url):
    """
    From given URL, retrieve the relative path (URL without domain and starting
    slash).

    Arguments:
        site_url (string): Website URL to remove from given ``url``
            argument.
        url (string): Full URL (starting with http/https) to make relative to
            website URL.

    Returns:
        string: Admin change view URL path for given model object.
    """
    if url.startswith(site_url):
        return url[len(site_url):]

    return url


def get_admin_add_url(model):
    """
    Return the right admin URL for add form view for given class.

    Arguments:
        model (Model object): A model object to use to find its admin
            add form view URL.

    Returns:
        string: Admin add form view URL path.
    """
    url_pattern = "admin:{app}_{model}_add"

    return reverse(url_pattern.format(
        app=model._meta.app_label,
        model=model._meta.model_name
    ))


def get_admin_change_url(obj):
    """
    Return the right admin URL for a change view for given object.

    Arguments:
        obj (Model object): A model object instance to use to find its admin
            change view URL.

    Returns:
        string: Admin change view URL path.
    """
    url_pattern = "admin:{app}_{model}_change"

    return reverse(url_pattern.format(
        app=obj._meta.app_label,
        model=obj._meta.model_name
    ), args=[
        obj.pk
    ])


def get_admin_list_url(model):
    """
    Return the right admin URL for a list view for given class.

    Arguments:
        model (Model object): A model object to use to find its admin
            list view URL.

    Returns:
        string: Admin list view URL path.
    """
    url_pattern = "admin:{app}_{model}_changelist"

    return reverse(url_pattern.format(
        app=model._meta.app_label,
        model=model._meta.model_name
    ))


def decode_response_or_string(content):
    """
    Shortand to get HTML string from either a TemplateResponse (as returned
    from Django test client) or a simple string so you can blindly give a
    response or a string without to care about content type.

    Arguments:
        content (TemplateResponse or string): If content is a string it will
            just return it. If content is a TemplateResponse it will decode byte
            string from its ``content`` attribute.

    Returns:
        string: HTML string.
    """
    if isinstance(content, TemplateResponse):
        return content.content.decode()
    return content


def html_element(content):
    """
    Shortand to use Django HTML parsing on given content.

    This is more useful for comparaison on HTML parts.

    Arguments:
        content (TemplateResponse or string): HTML content to parse.

    Returns:
        django.test.html.Element: A Python object structure able to perform
        comparaison on a semantical way. See ``django.test.html.parse_html`` for
        more details.
    """
    return parse_html(
        decode_response_or_string(content)
    )


def html_pyquery(content):
    """
    Shortand to use Pyquery parsing on given content.

    This is more useful to dig in advanced HTML content. PyQuery is basically a
    wrapper around ``lxml.etree`` it helps with a more intuitive API (alike
    Jquery) to traverse elements but when reaching a node content it will
    return ``lxml.html.HtmlElement`` object which have a less intuitive API.

    Arguments:
        content (TemplateResponse or string): HTML content to parse.

    Returns:
        pyquery.PyQuery: A PyQuery object.
    """
    return pq(
        decode_response_or_string(content),
        parser='html'
    )


def queryset_values(queryset, names=["slug", "language"],
                    orders=["slug", "language"]):
    """
    An helper to just return a list of dict values ordered from given queryset.

    Arguments:
        queryset (Queryset): A queryset to turn to values.

    Keyword Arguments:
        names (list): A list of field names to return as values for each object.
            Default return "slug" and "language" values only.
        orders (list): A list of field names to order results.
            Default order first on "slug" then "language".

    Returns:
        list: A list of dict items for all result objects.
    """
    return list(
        queryset.values(*names).order_by(*orders)
    )


def compact_form_errors(form):
    """
    Build a compact dict of field errors without messages.

    This is a helper for errors, keeping it more easy to test since messages
    may be too long and can be translated which is more difficult to test.

    Arguments:
        form (django.forms.Form): A bounded form.

    Returns:
        dict: A dict of invalid fields, each item is indexed by field name and
        value is a list of error codes.
    """
    errors = {}

    for name, validationerror in form.errors.as_data().items():
        errors[name] = [item.code for item in validationerror]

    return errors


def build_post_data_from_object(model, obj, ignore=["id"]):
    """
    Build a payload suitable to a POST request from given object data.

    Arguments:
        model (django.db.models.Model): A model object used to find object
            attributes to extract values.
        obj (object): A instance of given model or a dict (like the one returned
            by a factory ``build()`` method.
        ignore (list): List of field name to ignore for value
            extraction. Default to "id" but will not be enough for any field
            with foreign keys, automatic primary keys, etc..

    Returns:
        dict: Payload data to use in POST request.
    """
    data = {}

    fields = [
        f.name for f in model._meta.get_fields()
        if f.name not in ignore
    ]

    for name in fields:
        if obj is dict:
            data[name] = obj.get(name)
        else:
            data[name] = getattr(obj, name)

    return data


def get_test_source(storage, *args, **kwargs):
    """
    Shortcut helper to create a dummy image, save it to the FS and get its
    final path from storage.

    Arguments:
        storage (django.core.files.storage.Storage): Storage class to use to
            save file to the file system.

    Keyword Arguments:
        storage (django.core.files.storage.Storage): Storage class to use to
            save file to the file system. This is a required argument.
        destination_dir (string): relative directory path where to save file
            in storage directory. Default to "pil".
        *args: Any other positionnal arguments are passed to
            ``create_image_file``.
        **kwargs: Any other Keyword arguments are passed to
            ``create_image_file``.

    Returns:
        django.core.files.File: File object with the right final file path.
    """
    destination_dir = kwargs.pop("destination_dir", "pil")

    image = create_image_file(*args, **kwargs)

    destination = Path(destination_dir) / Path(image.name)
    source_filepath = storage.save(destination, image)

    # Trick to update name to final destination since File object is not
    # automatically updated to the final filepath during storage.
    image.name = source_filepath

    return image



class SmartImageBaseException(Exception):
    """
    Base for every SmartFormat template tag exceptions.
    """
    pass


class InvalidFormatError(SmartImageBaseException):
    """
    Exception to be raised from smart_format template tag when given format is
    not invalid.
    """
    pass


class IncompatibleSvgToBitmap(SmartImageBaseException):
    """
    Exception to be raised from smart_format template tag when required format
    is a Bitmap formt but the source is a SVG.
    """
    pass


class IncompatibleBitmapToSvg(SmartImageBaseException):
    """
    Exception to be raised from smart_format template tag when required format
    is a SVG but the source is a Bitmap.
    """
    pass



import datetime

import uuid

from pathlib import Path


def get_unique_filename(filename):
    """
    Generate a new unique filename keeping file extension unchanged.

    Uniqueness is guaranteed by UUID.

    Arguments:
        filename (string): Original filename.

    Returns:
        pathlib.Path: Path for new filename.
    """
    return Path(
        str(uuid.uuid4())
    ).with_suffix(
        "".join(Path(filename).suffixes)
    )


def uploadto_unique(dirname, instance, filename):
    """
    Return a filepath with date patterns formatted and unique filename.

    So an usage like this: ::

        def file_uploadto(instance, filename):
            return uploadto_unique("ping/%Y/%m/%d", instance, filename))

        class Foo(models.Model):
            file = models.FileField("A file", upload_to=file_uploadto)

    Would save a file ``foo.jpg`` to a filepath like: ::

        ping/2012/10/15/16fd2706-8baf-433b-82eb-8c7fada847da.jpg

    Arguments:
        dirname (string): Directory path for storage, can include date patterns.
        instance (object): Model object instance.
        filename (string): Original filename.

    Returns:
        pathlib.Path: A Path object for the filepath to save.
    """
    dirname = Path(
        datetime.datetime.now().strftime(str(dirname))
    )

    return dirname / get_unique_filename(filename)



import io

from PIL import Image as PILimage

from django.core.files import File


def create_image_file(filename=None, size=(100, 100), color="blue",
                      format_name="PNG"):
    """
    Return a File object with a dummy generated image on the fly by PIL or
    possibly a SVG file.

    With default argument values the generated image will be a simple blue
    square in PNG.

    Keyword Arguments:
        filename (string): Filename for created file, default to ``color``
            value joined to extension with ``format`` value in lowercase (or
            ``jpg`` if format is ``JPEG``).
            Note than final filename may be different if all tests use the same
            one since Django will append a hash for uniqueness.
        format_name (string): Format name as available from PIL: ``JPEG``,
            ``PNG`` or ``GIF``. ``SVG`` format is also possible to create a
            dummy SVG file.
        size (tuple): A tuple of two integers respectively for width and height.
        color (string): Color value to fill image, this should be a valid value
            for ``PIL.ImageColor``:
            https://pillow.readthedocs.io/en/stable/reference/ImageColor.html
            or a valid HTML color name for SVG format.

    Returns:
        django.core.files.File: File object.
    """
    ext = format_name.lower()
    # Enforce correct file extension depending format
    if ext == "jpeg":
        ext = "jpg"

    filename = filename or "{}.{}".format(color, ext)

    # Manage correct mode depending format
    mode = "RGB"
    if format_name == "PNG":
        mode = "RGBA"

    # Create a SVG file
    if format_name == 'SVG':
        width, height = size
        html = (
            """<svg xmlns="http://www.w3.org/2000/svg" """
            """viewBox="0 0 {width} {height}">"""
            """<path fill="{color}" d="M0 0h{width}v{height}H0z"/>"""
            """</svg>"""
        ).format(
            width=str(width),
            height=str(height),
            color=color,
        )
        thumb_io = io.StringIO(html)
    # Create an image file for every other formats
    else:
        thumb = PILimage.new(mode, size, color)
        thumb_io = io.BytesIO()
        thumb.save(thumb_io, format=format_name)

    return File(thumb_io, name=filename)



from django.conf import settings


class SmartFormatMixin:
    """
    A mixin to inherit from a model so it will have a helper method to manage image
    formats.

    Example:

        You just have to inherit this class then define a method dedicated to your
        field: ::

            from django.db import models
            from smart_media.modelfields import SmartMediaField

            class MyModel(SmartFormatMixin, models.Model):
                media = SmartMediaField(
                    "media",
                    max_length=255,
                    null=True,
                    blank=True,
                    default=None,
                    upload_to="sample/media/%y/%m",
                )

                def get_media_format(self):
                    return self.media_format(self.media)
    """

    def media_format(self, mediafile):
        """
        Common method to perform a naive check about image format using file
        extension.

        Retrieved format will come from setting ``SMART_FORMAT_AVAILABLE_FORMATS``.

        Obviously since it use the file extension, found format is not to be
        100% trusted. You may think about validate your files strictly if strong
        security is a matter for you and file upload is open to untrusted user.

        Arguments:
            mediafile (object): Either a FileField, ImageField or any other
                object which implement a ``name`` attribute which return the
                filename.

        Return:
            string: Format name if filename extension match to any available
            format extension from available formats else it returns ``None``.
        """
        if mediafile:
            ext = mediafile.name.split(".")[-1].lower()

            for fileext, formatname in settings.SMART_FORMAT_AVAILABLE_FORMATS:
                if ext == fileext:
                    return formatname

        return None



from django.conf import settings
from django.core.validators import FileExtensionValidator
from django.forms.fields import FileField, ImageField

from .widgets import ClearableFileInputButton


class SmartMediaField(FileField):
    """
    A FileField which define the widget ``ClearableFileInputButton`` and default
    validator ``FileExtensionValidator`` for allowed extensions (as defined from
    setting ``SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS``).
    """
    widget = ClearableFileInputButton
    default_validators = [
        FileExtensionValidator(
            allowed_extensions=settings.SMARTIMAGE_ALLOWED_IMAGE_EXTENSIONS
        ),
    ]


class SmartImageField(ImageField):
    """
    An ImageField which define the widget ``ClearableFileInputButton``, the default
    validator is still the one from ImageField and so the field won't allow for SVG
    file.
    """
    widget = ClearableFileInputButton



from pathlib import Path

from django.core.exceptions import FieldDoesNotExist


def auto_purge_files_on_delete(fieldnames):
    """
    Return a callable to set as a signal receiver to purge related files from deleted
    object.

    When connected to your model, just before an object is deleted the signal check
    defined fields for their files and remove them from your Filesystem so you won't
    keep files from object that do not exists anymore.

    This is to be used with signal ``django.db.models.signals.post_delete`` and since
    this work with shallow function you will need to set ``weak=False`` on signal
    method ``connect``.

    Example:

        Commonly you connect signal in your models file and define a list of fields to
        watch for purge: ::

            from django.db import models
            from django.db.models.signals import post_delete
            from smart_media.signals import auto_purge_files_on_delete

            class MyModel(models.Model):
                cover = models.FileField(...)

            post_delete.connect(
                auto_purge_files_on_delete(["cover"]),
                dispatch_uid="mymodels_files_on_delete",
                sender=MyModel,
                weak=False,
            )

    Remember to set an unique key name to argument ``dispatch_uid``.

    Arguments:
        fieldnames (list): List of model field names to seek for file to remove. If
            a given field name does not exists for given model instance an exception
            is raised. It is important to keep this list up to date with your model.

    Returns
        callable: A callable function with expected signature from signal emitter.
    """
    def _receiver_func(sender, instance, **kwargs):
        unknow_fields = [
            fieldname
            for fieldname in fieldnames
            if not hasattr(instance, fieldname)
        ]

        if unknow_fields:
            raise FieldDoesNotExist((
                "Function 'auto_purge_files_on_delete' was given invalid field names"
                " for model '{model}': {fields}"
            ).format(
                fields=", ".join(unknow_fields),
                model=instance.__class__.__name__,
            ))

        for fieldname in fieldnames:
            field = getattr(instance, fieldname, None)
            if field and Path(field.path).is_file():
                field.storage.delete(field.path)

    return _receiver_func


def auto_purge_files_on_change(fieldnames):
    """
    Return a callable to set as a signal receiver to purge old related files for a
    modified object with new uploaded files.

    When connected to your model, just before an existing object is saved the signal
    check defined fields for their current files, if a current file is different than
    the one from the changes, it is removed from your Filesystem so you won't
    keep files that have been deleted or changed.

    It is safe if old file does not exists anymore.

    This is to be used with signal ``django.db.models.signals.pre_save`` and since
    this work with shallow function you will need to set ``weak=False`` on signal
    method ``connect``.

    This receiver perform an additional get queryset on object to get its previous
    value just before current save.

    Example:

        Commonly you connect signal in your models file and define a list of fields to
        watch for purge: ::

            from django.db import models
            from django.db.models.signals import pre_save
            from smart_media.signals import auto_purge_files_on_change

            class MyModel(models.Model):
                cover = models.FileField(...)

            pre_save.connect(
                auto_purge_files_on_change(["cover"]),
                dispatch_uid="mymodels_files_on_change",
                sender=MyModel,
                weak=False,
            )

    Remember to set an unique key name to argument ``dispatch_uid``.

    Arguments:
        fieldnames (list): List of model field names to seek for file to remove. If
            a given field name does not exists for given model instance an exception
            is raised. It is important to keep this list up to date with your model.

    Returns
        callable: A callable function with expected signature from signal emitter.
    """
    def _receiver_func(sender, instance, **kwargs):
        unknow_fields = [
            fieldname
            for fieldname in fieldnames
            if not hasattr(instance, fieldname)
        ]

        if unknow_fields:
            raise FieldDoesNotExist((
                "Function 'auto_purge_files_on_change' was given invalid field names"
                " for model '{model}': {fields}"
            ).format(
                fields=", ".join(unknow_fields),
                model=instance.__class__.__name__,
            ))

        # Only run for object which have already been saved
        if not instance.pk:
            return False

        # Get the old saved version object
        try:
            old_obj = sender.objects.get(pk=instance.pk)
        except sender.DoesNotExist:
            return False

        # Store old files
        old_fieldvalues = {
            fieldname: getattr(old_obj, fieldname, None)
            for fieldname in fieldnames
        }

        # Clean file when field value changed to a new file and if there was an old
        # existing file
        for fieldname in fieldnames:
            field = getattr(instance, fieldname, None)

            if (
                old_fieldvalues[fieldname] and
                old_fieldvalues[fieldname] != field and
                Path(old_fieldvalues[fieldname].path).is_file()
            ):
                field.storage.delete(old_fieldvalues[fieldname].path)

    return _receiver_func



from django import template
from django.conf import settings

from sorl.thumbnail.shortcuts import get_thumbnail

from ..exceptions import (
    InvalidFormatError, IncompatibleSvgToBitmap, IncompatibleBitmapToSvg
)
from ..mixins import SmartFormatMixin
from ..thumbnailing import SvgFile

register = template.Library()


@register.simple_tag
def media_thumb(source, geometry, *args, **kwargs):
    """
    Determine the right format and return the Sorl thumb file url path for
    given image file.

    Arguments:
        source (object): Either a FileField or an ImageField.
        geometry (string): Geometry string as expected by Sorl. It should be
            something like ``200`` for 200px width and automatic height or
            ``200x100`` for 200px width and 100px height.
        *args: Not used, there is no other expected positionnal arguments.
        **kwargs: Keyword arguments are passed to Sorl, watch its documentation
            for more details.

    Keyword Arguments:
        format (string):
            An available format name from settings ``SMART_FORMAT_AVAILABLE_FORMATS``:

            * ``PNG``, ``JPEG`` and ``GIF`` enforce the thumb format no matter
              the source image format. Be careful to not enforce a format where
              a SVG is a possible source format, since Sorl can't read SVG.
            * ``SVG`` will not produce any thumb, just return the same path
              than the source one since SVG does not need a thumb and is not
              supported from image libraries.
            * ``auto`` to automatically find and use the same format than
              the source image. This is the recommended way.

            When argument is empty the default value is ``auto``.

    Example:
        The most basic usage is to define only thumbnail geometry in default
        "auto" format mode: ::

            {% load smart_format %}
            {% media_thumb instance.image "250x200" as thumb %}
            <img src="{{ thumb.url }}" alt="">

        You can also enforce a specific format: ::

            {% load smart_format %}
            {% media_thumb instance.image "250x200" format="JPEG" as thumb %}
            <img src="{{ thumb.url }}" alt="">

        Keep it in mind this will raise an error with SVG image files (if you
        allowed SVG format in plugins) since image libraries (like Pillow or
        ImageMagick) has no support for SVG.

        Every argument but ``format`` is passed to Sorl, so for cropping an image you
        can do : ::

            {% load smart_format %}
            {% media_thumb instance.image "250x200" crop="center" as thumb %}
            <img src="{{ thumb.url }}" alt="">

    Return:
        sorl.thumbnail.images.ImageFile or SvgFile: Either Sorl ImageFile instance for
        created thumb or SvgFile which mimic Sorl ImageFile.
    """
    # Safe return when there is no source file
    # NOTE: Should trigger a warning log ?
    if not source:
        return None

    required_format = kwargs.pop("format", "auto")
    source_extension = source.name.split(".")[-1].lower()

    available_extensions = [k for k, v in settings.SMART_FORMAT_AVAILABLE_FORMATS]

    # Automatic format guess
    if required_format.lower() == "auto":
        required_format = SmartFormatMixin().media_format(source)
    # Specific SVG format
    elif required_format.lower() == "svg":
        # A bitmap file cannot be converted to a SVG
        if source_extension != "svg":
            msg = ("Incompatible required format (SVG) with source format "
                   "(Bitmap).")
            raise IncompatibleBitmapToSvg(msg.format(
                format_name=required_format.lower(),
            ))
        # SVG to SVG is correct (nothing to do)
        return SvgFile(source)
    # Unknown format
    elif required_format.lower() not in available_extensions:
        msg = ("Required format '{format_name}' does not match available "
               "formats: {available}")
        raise InvalidFormatError(msg.format(
            format_name=required_format.lower(),
            available=", ".join(available_extensions),
        ))
    # SVG source with required bitmap format
    elif source_extension == "svg":
        msg = ("Incompatible required format (Bitmap) with source format "
               "(SVG).")
        raise IncompatibleSvgToBitmap(msg.format(
            format_name=required_format.lower(),
        ))

    kwargs["format"] = required_format

    # Vector format does not create a thumb and so just return original
    # source file
    if required_format == "SVG":
        return SvgFile(source)

    # Here we assume the format is ok, if it has been manually defined, user
    # is responsible about possible error with its content
    return get_thumbnail(source, geometry, **kwargs)



from django.core.files.utils import validate_file_name
from django.db import models

from .fields import SmartMediaField as SmartMediaFormField
from .utils.file import uploadto_unique
from .validators import SmartMediaFileExtensionValidator


class SmartMediaField(models.FileField):
    """
    A model file field with every "smart" features.

    * Set the ``SmartMediaFileExtensionValidator`` validator;
    * Use the form field ``SmartMediaField`` to get the proper "smart" clearable file
      widget;
    * Generate filename as an unique UUID with ``uploadto_unique`` function;

    Example:

        Usage is identical to a ``models.FileField``: ::

            cover = SmartMediaField(
                "cover",
                max_length=255,
                null=True,
                blank=True,
                default=None,
                upload_to="sample/cover/%y/%m",
            )

    .. Note::

        If attribute ``upload_to`` value is a callable then the UUID filename behavior
        is disabled.

    .. Note::

        Sadly this just won't work on its own in Django admin which makes some internal
        resolutions between model fields and widgets. So you will need to define a rule
        to get the right widget for this field, see
        ``smart_media.admin.SmartModelAdmin``.

    .. Warning::

        Django won't be able to detect change on settings. So you will have to create a
        custom data migration on your own if you want to clean files with previously
        allowed extensions but removed from setting.

    """
    default_validators = [
        SmartMediaFileExtensionValidator(),
    ]

    def generate_filename(self, instance, filename):
        """
        Override ``FileField.generate_filename()`` to change original uploaded filename
        to an UUID.

        'upload_to' date pattern is still properly formatted and the UUID behavior is
        ignored if 'upload_to' is a callable.
        """
        if callable(self.upload_to):
            filename = self.upload_to(instance, filename)
        else:
            filename = uploadto_unique(str(self.upload_to), instance, filename)

        filename = validate_file_name(filename, allow_relative_path=True)

        return self.storage.generate_filename(filename)

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": SmartMediaFormField,
                **kwargs,
            }
        )



from django.forms.widgets import FileInput, ClearableFileInput


class FileInputButtonBase(FileInput):
    """
    A simple FileInput override to use a custom template.

    The base version does not include the required medias so you can include
    them yourself in your form controler, this is a good idea in some
    environment like DjangoCMS which needs to use a lot of ``!important`` marks
    to override the admin stylesheets and is not included from this widget stylesheets.

    Default behavior if no ``class`` attribute is given is to apply the CSS
    classname ``fileinputbutton__input`` to the input.

    .. Note::
        When your input file enable the ``multiple`` attribute to accept more
        than one file, label input will be rendered to a default string like
        "4 files" which is not translatable (this is done in the Javascript
        source).

        If you want a translatable string just pass it to widget instance
        attribute ``data-multiple-caption`` like: ::

            FileInputButton(attrs={
                "data-multiple-caption": gettext_lazy("{count} files selected")
            })

        Pattern ``{count}`` will be replaced with the number of selected files.
    """
    template_name = "smart_image/fileinputbutton_basic.html"

    def __init__(self, attrs=None):
        default_attrs = {'class': 'fileinputbutton__input'}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)


class ClearableFileInputButtonBase(ClearableFileInput):
    """
    Alike ``FileInputButtonBase`` but to extend with the additional clear checkbox.

    This won't support the "multiple" input file feature.
    """
    template_name = "smart_image/fileinputbutton_clearable.html"

    def __init__(self, attrs=None):
        default_attrs = {'class': 'fileinputbutton__input'}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(default_attrs)


class FileInputButton(FileInputButtonBase):
    """
    ``FileInputButtonBase`` version which includes the required assets to customize
    its layout and turn the file input as a button.
    """
    class Media:
        css = {
            "all": ("smart_image/css/fileinputbutton.css",),
        }
        js = ("smart_image/js/fileinputbutton.js",)


class ClearableFileInputButton(ClearableFileInputButtonBase):
    """
    ``ClearableFileInputButtonBase`` version which includes the required assets to
    customize its layout and turn the file input as a button.
    """
    class Media:
        css = {
            "all": ("smart_image/css/fileinputbutton.css",),
        }
        js = ("smart_image/js/fileinputbutton.js",)



from django.contrib import admin as django_admin

from .modelfields import SmartMediaField
from .widgets import ClearableFileInputButton


class SmartModelAdmin(django_admin.ModelAdmin):
    """
    A model admin class to include the right widget definition for model field
    ``modelfields.SmartMediaField``.

    This is required because ModelAdmin translate model field to widgets using a list
    of harcoded relations between fields and widgets. It results field definition in
    ``SmartMediaField.formfield`` is ignored and translate ``SmartMediaField`` to
    ``AdminFileWidget`` instead of expected ``ClearableFileInputButton``.

    So either your model admin inherits from ``SmartModelAdmin`` or simply copy its
    ``SmartModelAdmin.formfield_overrides`` into it.

    It is only required with model forms in Django admin.

    Example:

        Your model admin just have to inherit from this class: ::

            from django.contrib import admin
            from smart_media.admin import SmartModelAdmin
            from myapp.models import MyModel

            @admin.register(MyModel)
            class MyModelAdmin(SmartModelAdmin):
                ...

    Or copy the source of ``SmartModelAdmin.formfield_overrides`` (see source) into
    your own model admin.
    """
    formfield_overrides = {
        SmartMediaField: {
            "widget": ClearableFileInputButton,
        },
    }
